#!/usr/bin/perl -w
#
# $Id: imap2mbox.pl,v 1.7 2004/08/29 11:28:13 norgaard Exp $
#
use strict;
use warnings;
#
use Getopt::Std;
use IMAP::Admin;                # To set the privileges
use Mail::IMAPClient;           # To migrate mailbox

$| = 1;

my ($b,$f,$m,$s,$t) = (0,0,0,0,0);   # Total mails transfered
my ($box,%opts);
# Initializing:

getopt('hvp:s:o:u:',\%opts);

if ($opts{h}) { # Help text
  print <<'END_OF_HELP';
Usage: imap2mbox.pl [-u <admin> -s <server> -p <passwd> <mailfolder> ... ]
See 'perldoc imap2mbox.pl' for more details.
END_OF_HELP
  exit 0;
}

my $sysmail = "From MAILER-DAEMON " . scalar localtime() . "\n"
     . "From: Mail System Internal Data <MAILER-DAEMON\@localhost>\n"
     . "Subject: DON'T DELETE THIS MESSAGE -- FOLDER INTERNAL DATA\n"
     . "Date: " . scalar gmtime() . "\n"
     . "Message-ID: <" . time() . "\@localhost>\n"
     . "X-IMAP: " . time() . " 0000000000\n"
     . "Status: RO\n"
     . "
This text is part of the internal format of your mail folder, and is not
a real message.  It is created automatically by the mail system software.
If deleted, important folder data will be lost, and it will be re-created
with the data reset to initial values.\n\n";

my $debug  = $opts{'d'};
my $outdir = $opts{'o'} ? $opts{'o'} : '/tmp';
my $server = $opts{'s'} ? $opts{'s'} : 'localhost';
my $admin  = $opts{'u'} ? $opts{'u'} : 'cyrus';
my $passwd = $opts{'p'} ? $opts{'p'} : '';

if (!$passwd) {
  print "Enter administrator password for $admin: ";
  $passwd = <STDIN>;
  chomp $passwd;
}

#####################################################################
#
# Connect to IMAP server as admin

my $imap = IMAP::Admin->new(
    'Server'   => $server,
    'Login'    => $admin,
    'Password' => $passwd,
    'CRAM'     => 2,
    'Debug'    => $debug)
  or die "Could not authenticate: $@\n";

# Get list of folders to transfer
if (@ARGV) {
  foreach my $arg (@ARGV) {
    my ($user, $domain) = $arg =~ /^([^\@]+)(\@.*|)/;

    foreach my $folder ($imap->list("user.$user$domain"), $imap->list("user.$user.*$domain")) {
      $box->{$folder} = "$outdir/$folder";
    }
  }
}
else {
  foreach my $folder ($imap->list("*")) {
    $box->{$folder} = "$outdir/$folder";
  };
};

# Set all permisions for admin on folders
print "Setting permissions: "; # All permissions!
foreach my $folder (sort keys %{$box}) {
  $imap->set_acl("$folder", $admin, "lrswipcda")
  and warn "failed setting permissions on $folder: $@\n" and next;
}
print "OK\n";

#####################################################################
#
# Transfer folders

my $start = time();

my $client = Mail::IMAPClient->new(
    'Server'       => $server,
    'User'         => $admin,
    'Password'     => $passwd,
    'Debug'        => $debug )
  or die "Failed connect to $server. $@";

foreach my $folder (keys %{$box}) {
  print "Transfering messages from $folder to $box->{$folder}\n";
  my ($h,$i,$j,$k) = imap2mbox($client, $folder,$box->{$folder});
  $s += -s $box->{$folder};
  $b += $h; # total bytes transfered
  $f += $i; # total messages failed
  $m += $j; # total messages
  $t += $k; # total messages transfered
};

$client->disconnect;

print "Transfered $t mails ($b bytes) of $m total, $f failed. $s bytes written\n";
print "Processed: ", $m/(time()-$start), " mails/second\n";
print "Transfered:", $t/(time()-$start), " mails/second, ", $b/(time()-$start), " bytes/second\n";
print "Time: ", time()-$start, " seconds\n";


#####################################################################
#
# Revoke admins permissions

print "Revoking admin's permissions: "; # All permissions!
foreach my $folder (keys %{ $box }) {
  $imap->delete_acl("$folder",$admin) and
    warn "failed revoke on $folder: $@\n" and next;
}
print "OK\n";

$imap->close;

#####################################################################
#
# Subrutines

sub imap2mbox {
  my ($client, $folder, $mbox) = @_;
  my ($b,$f,$m,$s,$t) = (0,0,0,0,0);

  $client->select($folder) or
    warn "Could not select: $@\n" and return ($b,$f,$m,$t);

  open(FLD,">","$mbox");

  print FLD $sysmail;

  foreach my $msgid ($client->messages) {
    $m++;
    print "Transfering message $msgid: " if $debug;
    my $msg = $client->message_string($msgid)
      or (warn "Failed. $@\n" and $f++ and next);
    my ($header,$body) = split /\r?\n\r?\n/, $msg, 2;

    if ($header =~ /^X-IMAP:\s*\d+\s*\d+/igmo) {
      print $header if $debug;
      $s++, next;                                        # Mbox system mail
    }

    $b += do { use bytes; length($msg); };

    $msg =~ s/\r//og;               # Remove carriage-return character
    $msg =~ s/\n\nFrom /\n\n> From /igo;

    # Print mailbox format separator line
    print FLD "From imap2mbox\@localhost ", scalar localtime(), "\n";
    print FLD $msg, "\n\n";
    print "OK\n" if $debug;
    $t++;
  };

  close(FLD);
  print "Transfered $t messages ($b bytes) of $m total, $f failed, $s skipped.\n";
  return ($b,$f,$m,$t);
};

#####################################################################
#
# Stub documentation

=head1 imap2mbox.pl

=head2 SYNOPSIS

  imap2mbox.pl [ options ]

=head2 DESCRIPTION

This script is intended for the administrator for taking backup of mail stored
on an IMAP server or migrating from an IMAP server to standard Unix mailbox.
The standard mailbox format is mostly usefull as it allows reading mail
independend of the imap server.

=head2 USAGE

The script takes a number of command line options, and arguments. The
arguments not associated with an option are account names to be transfered.
If no accounts are specified, all are transfered. The output is stored in
files with the same name as the original folder.

=head2 OPTIONS

=item -h print help text

=item -o output directory where the mailboxes will be stored

=item -p user password, if not specified the script will prompt for a password
NOTE: Specifying password on the command line may store your password in clear
text in the shell history.

=item -s imap server, default is 'localhost'

=item -u imap administrator. If this option is left out, the user is prompted.
Default is 'cyrus'.

=item -v verbose/debug

=head2 VIRTUAL DOMAINS

The script supports virtual domains. If no accounts are specified, all are
transfered, including those belonging to a virtual domain. Individual accounts
from a virtual domain can be transfered by giving the full realm.

=head2 SKIPPED EMAILS

The script skips the system e-mail common to the mailbox format, should it
have been transfered. A custom e-mail is inserted in the mailbox file for
the same purpose.

=head2 BUGS AND KNOWN PROBLEMS

None

=head2 DEPENDENCIES

   IMAP::Admin
   Mail::IMAPClient

=head2 COPYRIGHT

Copyright (C) 2004 Erik Norgaard. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

DISCLAIMER:

THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ''AS IS''. ANY 
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED.

=cut
