#!/usr/bin/perl -w
#
# $Id: user_imap2mbox.pl,v 1.3 2004/08/29 11:28:13 norgaard Exp $
#
use strict;
use warnings;
#
use Getopt::Std;
use Mail::IMAPClient;           # To migrate mailbox

$| = 1;

my ($b,$f,$m,$s,$t) = (0,0,0,0,0);   # Total mails transfered
my ($box,@folders,%opts);
# Initializing:

getopt('hvp:s:o:u:',\%opts);

if ($opts{h}) { # Help text
  print <<'END_OF_HELP';
Usage: imap2mbox.pl [-u <user> -s <server> -p <passwd> <mailfolder> ... ]
See 'perldoc imap2mbox.pl' for more details.
END_OF_HELP
  exit 0;
}

my $sysmail = "From MAILER-DAEMON " . scalar localtime() . "\n"
     . "From: Mail System Internal Data <MAILER-DAEMON\@localhost>\n"
     . "Subject: DON'T DELETE THIS MESSAGE -- FOLDER INTERNAL DATA\n"
     . "Date: " . scalar gmtime() . "\n"
     . "Message-ID: <" . time() . "\@localhost>\n"
     . "X-IMAP: " . time() . " 0000000000\n"
     . "Status: RO\n"
     . "
This text is part of the internal format of your mail folder, and is not
a real message.  It is created automatically by the mail system software.
If deleted, important folder data will be lost, and it will be re-created
with the data reset to initial values.\n\n";

my $debug  = $opts{'d'};
my $outdir = $opts{'o'} ? $opts{'o'} : '/tmp';
my $server = $opts{'s'} ? $opts{'s'} : 'localhost';
my $user   = $opts{'u'} ? $opts{'u'} : '';
my $passwd = $opts{'p'} ? $opts{'p'} : '';

if (!$user) {
  print "Enter username: ";
  $user = <STDIN>;
  chomp $user;
}

if (!$passwd) {
  print "Enter password for $user: ";
  $passwd = <STDIN>;
  chomp $passwd;
}

#####################################################################
#
# Connect to IMAP server as admin

my $client = Mail::IMAPClient->new(
    'Server'       => $server,
    'User'         => $user,
    'Password'     => $passwd,
    'Debug'        => $debug )
  or die "Failed connect to $server. $@";

if (@ARGV) {
  @folders = @ARGV;
}
else {
  @folders = $client->folders();
}

#####################################################################
#
# Transfer folders

my $start = time();

foreach my $folder (@folders) {
  my $mbox = "$outdir/$folder";
  print "Transfering messages from $folder to $mbox\n";
  my ($h,$i,$j,$k) = imap2mbox($client, $folder, $mbox);
  $s += -s $mbox;
  $b += $h; # total bytes transfered
  $f += $i; # total messages failed
  $m += $j; # total messages
  $t += $k; # total messages transfered
};

$client->disconnect;

print "Finnished processing folders. Summary:\n";

print "Transfered $t mails/$b bytes of $m total. $s bytes written.\n";
print "Failed transfering $f mails, ", 100*$f/$t, "% errors\n";
print "Processed: ", $m/(time()-$start), " mails/second\n";
print "Transfered:", $t/(time()-$start), " mails/second, ", $b/(time()-$start), " bytes/second\n";
print "Time: ", time()-$start, " seconds\n";

#####################################################################
#
# Subrutines

sub imap2mbox {
  my ($client, $folder, $mbox) = @_;
  my ($b,$f,$m,$s,$t) = (0,0,0,0,0);

  $client->select($folder) or
    warn "Could not select: $@\n" and return ($b,$f,$m,$t);

  open(FLD,">","$mbox");

  print FLD $sysmail;

  foreach my $msgid ($client->messages) {
    $m++;
    print "Transfering message $msgid: " if $debug;
    my $msg = $client->message_string($msgid)
      or (warn "Failed. $@\n" and $f++ and next);
    my ($header,$body) = split /\r?\n\r?\n/, $msg, 2;

    if ($header =~ /^X-IMAP:\s*\d+\s*\d+/igmo) {
      print $header if $debug;
      $s++, next;                                        # Mbox system mail
    }

    $b += do { use bytes; length($msg); };

    $msg =~ s/\r//og;               # Remove carriage-return character
    $msg =~ s/\n\nFrom /\n\n> From /igo;

    # Print mailbox format separator line
    print FLD "From imap2mbox\@localhost ", scalar localtime(), "\n";
    print FLD $msg, "\n\n";
    print "OK\n" if $debug;
    $t++;
  }

  close(FLD);
  print "Transfered $t messages ($b bytes) of $m total, $f failed, $s skipped.\n";
  return ($b,$f,$m,$t);
};

#####################################################################
#
# Stub documentation

=head1 user_imap2mbox.pl

=head2 SYNOPSIS

  user_imap2mbox.pl [ options ]

=head2 DESCRIPTION

This script allows the user to transfer all or some mail folders from an
IMAP server and store them in the classic Unix mailbox format. The script
is usefull for taking backup or offline mail reading.

In the Unix mailbox format all mails are placed in one file separated with
by a "\n\nFrom " line. Normally unix systems add envelope information to
this line. Envelope information is not present on the IMAP server, instead
the script adds a line informing that the mail was added by the script.

=head2 USAGE

The script takes a number of command line options, and arguments. The
arguments not associated with an option are foldernames to be transfered.
If no folders are specified, all are transfered. The output is stored in
files with the same name as the original folder.

=head2 OPTIONS

=item -h print help text

=item -o output directory where the mailboxes will be stored

=item -p user password, if not specified the script will prompt for a password
NOTE: Specifying password on the command line may store your password in clear
text in the shell history.

=item -s imap server, default is 'localhost'

=item -u imap account. Note, if this is an account in a virtual domain, the
full address must be specified. If this option is left out, the user is
prompted.

=item -v verbose/debug

=head2 VIRTUAL DOMAINS

The script transfers mailfolders from any type of account, to transfer mail
folders from a virtual domain, the full realm must be specified as user name.
Files are stored with the same name as the script sees.

=head2 SKIPPED EMAILS

The script skips the system e-mail common to the mailbox format, should it
have been transfered. A custom e-mail is inserted in the mailbox file for
the same purpose.

=head2 BUGS AND KNOWN PROBLEMS

None.

=head2 DEPENDENCIES

   Mail::IMAPClient

=head2 COPYRIGHT

Copyright (C) 2004 Erik Norgaard. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

DISCLAIMER:

THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ''AS IS''. ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.

=cut

