#!/usr/bin/perl -w
#
# $Id: user_mbox2imap.pl,v 1.6 2004/08/29 11:28:13 norgaard Exp $
#
# This script migrates unix-style mailbox to a (cyrus) imap server.
#
use strict;
use warnings;
#
use File::Basename;
use Getopt::Std;                # Command line options
use IMAP::Admin;                # To create mailbox
use Mail::IMAPClient;           # To migrate mailbox

$| = 1;

my $start = time();
my $box;
my ($f, $m, $t) = (0,0,0,0);    # Mails failed, parsed, transfered
my ($b, $s)     = (0,0);        # Bytes transfered, total size
my (%opts);
# Initializing:

getopts('hvp:s:u:',\%opts);

if ($opts{h}) { # Help text
  print <<'END_OF_HELP';
Usage: user_mbox2imap.pl [ -u <user> -s <server> -p <passwd> <glob> ... ]
See 'perldoc user_mbox2imap.pl' for more details.
END_OF_HELP
exit 0;
}
#@ARGV or die "No source file specified\n";

my $debug  = $opts{'v'};
my $server = $opts{'s'} ? $opts{'s'} : 'localhost';
my $user   = $opts{'u'} ? $opts{'u'} : getpwuid($<);
my $passwd = $opts{'p'} ? $opts{'p'} : '';

if (!$user) {
  print "Enter username: ";
  $user = <STDIN>;
  chomp $user;
}

if (!$passwd) {
  print "Enter password for $user: ";
  $passwd = <STDIN>;
  chomp $passwd;
}

($user, my $domain) = $user =~ /^([^\@]+)(\@.*|)/;
my $home = (getpwnam($user))[7];

if (@ARGV) {
  foreach my $glob (@ARGV) {
    foreach my $mbox (glob($glob)) {
      next if basename($mbox) =~ /^\./o;
      next if basename($mbox) =~ /\.msf$/o;
      print "Checking file: $mbox\n";

      my $folder = $mbox;
      $folder =~ s/^$home\/?//go;
      $folder =~ s/\//./go;     # Replace unix separator with imap separator
      $box->{$mbox} = $folder;
    };
  }
}
else {
  $box = getmbox($user);
}

$box or die "No matching files found\n";

#####################################################################
#
# Connect

my $client = Mail::IMAPClient->new(
    'Server'       => $server,
    'User'         => "$user$domain",
    'Password'     => $passwd,
    'Debug'        => $debug )
  or die "Failed connect to $server. $@";

#####################################################################
#
# Transfer mail

print "Transfering mail for $user$domain.\n";

foreach my $mbox (sort keys %{$box}) {
  print "Transfering messages from $mbox to $box->{$mbox}\n";
  my ($h,$i,$j,$k) = mbox2imap($client,$mbox,$box->{$mbox});
  $s += -s $mbox;
  $b += $h;
  $f += $i;
  $m += $j;
  $t += $k;
};

# Disconnect: DONE!
$client->disconnect;

print "Finnished processing files. Summary:\n";

print "Transfered $t mails/$b bytes of $m total/$s bytes.\n";
print "Failed transfering $f mails, ", 100*$f/$t, "% errors\n";
print "Processed: ", $m/(time()-$start), " mails/second, ", $s/(time()-$start), " bytes/second\n";
print "Transfered:", $t/(time()-$start), " mails/second, ", $b/(time()-$start), " bytes/second\n";
print "Time: ", time()-$start, " seconds\n";

#####################################################################
#
# Subrutines

sub getmbox {
  # This rutine takes a user name as input and returns a reference to a hash
  # containing the user's mailbox'es as keys and corresponding imap folders
  # as value. For virtual domains, the domain must be appended.
  my $user = shift @_;
  my $box;

  if (defined(scalar getpwnam($user))) { # REAL user
    my ($home) = (getpwnam($user))[7];

    -d $home or
      (warn "Homedir not found: $home\n" and return);

    # Locate the INBOX.
    if (-f "$home/Mailbox") {       # INBOX in homedir
      $box->{"$home/Mailbox"} = "INBOX";
    }
    elsif ( -f "/var/mail/$user") { # INBOX in maildir
      $box->{"/var/mail/$user"} = "INBOX";
    }
    else {                          # NO INBOX!
      warn "No main mailbox for $user\n";
      return;
    };

    # Squirrelmail by default stores other mail here (wu-imap)
    foreach my $squirrel (glob("$home/INBOX.*")) {
      next if -z $squirrel;  # Skip empty boxes
      next if $squirrel =~ /\.msf$/;
      $squirrel =~ /$home\/INBOX\.(.*)/;
      $box->{"$squirrel"} = "$1";
    }

    # PINE by default stores other mail here
    foreach my $pine (glob("$home/mail/*")) {
      my $base = basename($pine);   # folder name
      next if $base =~ /^\./;       # Skip hidden files/directories
      next if $base =~ /\.msf$/;
      $box->{"$pine"} = "$base";
    }
    return $box;
  }

  else { return };
  return $box;
}

sub mbox2imap {
  # This rutine takes an Mail::IMAPClient object, unix mailbox filename and
  # imap destination folder as input and extracts mail from the mailbox file
  # into the imap folder. Returns the number of mails succesfully transfered

  my ($client,$mbox,$folder) = @_;
  my $mail;                        # Mail: header and body
  my ($d,$f,$m,$s,$t) = (0,0,0,0,0); # duble, failed, msg no, skip, transfered
  my ($b) = 0;
  my (@new, @replied);

  if (!$client->exists($folder)) { # Attemt to create the folder if not found
    print "Creating folder $folder: ";
    $client->create($folder) or
      (print "Could not create $folder: $@\n" and return ($b,$f,$m,$t));
    print "OK\n";
  };

  # Return if mailbox is really a directory or empty
  return ($b,$f,$m,$t) if (-d $mbox or -z $mbox);

  # Select the destination folder. NOTE: There is no unselect, so failing
  # here mess up things, unless error caught!
  $client->select($folder) or
    (print "Could not select: $@\n" and return ($b,$f,$m,$t));

  { # The new line delimiter is set to the mailbox' mail separator (almost).
    # NOTE: The first line of every message will contain some trailing junk
    # to be removed.
    my $tmpbox = $tmpdir . "/" . basename($mbox);
    copy($mbox,$tmpbox);

    local $/ = "\n\nFrom ";
    open(MAIL, "<", $tmpbox) or
      (print "Failed opening $mbox: $!\n" and return ($b,$f,$m,$t));

    while (my $msg = <MAIL>) {
      $m++;

      chomp($msg);                   # This removes trailing "\n\nFrom "
      $msg =~ "X-Envelope: " . $msg; # Sets non standard header line

      if (!defined($msg) or $msg =~ /^[\n\s]*$/go) {
        print "$mbox: Message $m is empty (no header nor body)\n";
        $s++;
        next;
      };

      my ($header,$body) = split /\n\n/, $msg, 2;

      if (!defined($header) or $header =~ /^[\n\s]*$/go) { # Empty header
        print "$mbox: Message $m has no header.\n";
        print $msg if $debug;
        $s++;
        next;
      }

      if (!defined($body) or $body =~ /^[\n\s]*$/go) {     # Empty body
        $msg .= "\n.\n";                                   # Add a dot
      }

      $header =~ s/\n\s+//go;                              # Unwrap header
      if ($header !~ /^[^\s]+:/gmo) {                      # Invalid header
        print "$mbox: Message $m has invalid header.\n";
        print "$header\n" if $debug;
        $s++;
        next;
      };

      if ($header =~ /^X-IMAP:\s*\d+\s*\d+/igmo) {
        print $header if $debug;
        $s++, next;                                        # Mbox system mail
      }

      my ($msg_id) = $header =~ /^Message-id:\s*([^\n]*)/igmo;
      if (!$msg_id) {
        print "$mbox: Message $m has no message id\n";
        print "$header\n" if $debug;
      }

      # Check if message has been transfered before:
      if ($msg_id and my $id = $client->search(HEADER => 'Message-id' =>
                                   $client->Quote($msg_id))) {
        $d++;
        next;
      };

      $msg =~ s/\n\n> From /\n\nFrom /go;

#      print "Transfering message: ";
      if (my $uid = $client->append($folder,$msg)) {
#        print "$uid, OK\n";
        push @new, $uid;
        $b += do { use bytes; length($msg); };
        if ($header =~ /^X-Status: A/miog) {
          push @replied, $uid;
        }
        $t++;
      } else {
        $f++;
        print "$mbox: Message $m failed: $@\n";
      };
    };
    close(MAIL);
    unlink($tmpbox);
  };

  print "Transfered $t messages ($b bytes) of $m total, $f failed, $d dublicates, $s skipped.\n";

  # We mark mails as unseen. This is necesary only to force some mailers to
  # retrieve all mails.
  if (@new) {
    print "Unsetting 'seen' flag: ";
    $client->unset_flag('\Seen',@new) ?
      print "OK\n" : print "Failed\n";
  };
  if (@replied) {
    print "Setting 'answered' flag: ";
    $client->set_flag('\Answered',@replied) ?
      print "OK\n" : print "Failed\n";
  };

  return ($b,$f,$m,$t);
};

#####################################################################
#
# Stub documentation

=head1 user_mbox2imap.pl

=head2 SYNOPSIS

  user_mbox2imap.pl [ options ]

=head2 DESCRIPTION

user_mbox2imap.pl is yet another a script for transfering Berkely style
mailboxes to a (cyrus) imap server. It uses the imap protocol, hence it
should be posible to transfer mail to any imap-server, however the scripts
has been developed for cyrus-imap and only tested with cyrus-imap.

This script is intended to be run by the individual user for migrating
standard Unix mailboxes to the users imap folders. It may be run multiple
times, if a message shares message-id with a message allready in the
destination folder it is not transfered. Messages transfered all have the
'Seen' flag removed (this just appeared to work), and if there is a
"X-Status: R" entry in the header, the 'Answered' flag is set.

If the script is run without arguments the script will lookup mailfolders
in common places and place them in corresponding paths. The user may specify
a glob of files to be transfered. Then the mail is placed under a path
corresponding to the path in the glob.

=head2 USAGE

The script takes a number of command line options, and arguments. The
arguments not associated with an option are fileglobs matching the files
to be transfered, if no files are specified, the script will look for
user mailboxes in common places and transfer these.

=head2 OPTIONS

=item -h print help text

=item -p user password, if not specified the script will prompt for a password

=item -s imap server, default is 'localhost'

=item -u imap account. Note, if this is an account in a virtual domain, the
full address must be specified. If this option is left out, the user is
prompted.

=item -v verbose/debug

=head2 VIRTUAL DOMAINS

If the user's mail is to be transfered to a mailaccount in a virtual domain
simply specify the full address as username.

=head2 SKIPPED EMAILS

The script may skip some emails that would otherwise cause errors. Enabling
debugging with the -d debugging flag will cause headers of skipped mail to
be printed, along with other debugging information.

The mbox contains one mail for internal purposes, this is the first mail in
the file, and contains an X-IMAP field in the header. This mail is always
skipped. For this reason, you should expect one mail to be skipped in all
files.

=head2 BUGS AND KNOWN PROBLEMS

The script does not handle ".." in paths.

There is no support for recursion into subdirectories.

Mails with empty body has caused problems, a "." is added as body to these
emails (these e-mails are typically mails sent to list-servers containing
commands in the subject, hence no body).

The Mbox format splits mails with the pattern "\n\nFrom ", some mail clients
(PINE) does not escape this pattern prepending a "> " if occuring in the body
of the email. In such cases, only the part before the "\n\nFrom " will be
transfered and the remaining part will be treated as a new mail causing an
"invalid header" error.

=head2 DEPENDENCIES

   Mail::IMAPClient

=head2 COPYRIGHT

Copyright (C) 2004 Erik Norgaard. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

DISCLAIMER:

THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ''AS IS''. ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.

=cut
